use async_trait::async_trait;
use axum_login::{AuthUser, AuthnBackend, UserId};
use password_auth::verify_password;
use sea_orm::{ColumnTrait, DatabaseConnection, EntityTrait, QueryFilter};
use serde::{Deserialize, Serialize};
use thiserror::Error;
use tokio::task;

#[derive(Debug, Clone)]
pub struct UserDatabase {
    connection: DatabaseConnection,
}

impl UserDatabase {
    pub fn new(connection: DatabaseConnection) -> Self {
        UserDatabase {
            connection: connection,
        }
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct User {
    id: i64,
    pub username: String,
    password: String,
}

// Here we've implemented `Debug` manually to avoid accidentally logging the
// password hash.
impl std::fmt::Debug for User {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("User")
            .field("id", &self.id)
            .field("username", &self.username)
            .field("password", &"[redacted]")
            .finish()
    }
}

impl AuthUser for User {
    type Id = i64;

    fn id(&self) -> Self::Id {
        self.id
    }

    fn session_auth_hash(&self) -> &[u8] {
        self.password.as_bytes() // We use the password hash as the auth
                                 // hash--what this means
                                 // is when the user changes their password the
                                 // auth session becomes invalid.
    }
}

impl From<entity::user::Model> for User {
    fn from(value: entity::user::Model) -> Self {
        User {
            id: value.id,
            username: value.username,
            password: "".to_owned(), // TODO: password field.
        }
    }
}

// This allows us to extract the authentication fields from forms. We use this
// to authenticate requests with the backend.
#[derive(Debug, Clone, Deserialize)]
pub struct Credentials {
    pub username: String,
    pub password: String,
    pub next: Option<String>,
}

#[derive(Debug, Error)]
pub struct NotFoundError {
    pub details: String,
}

impl NotFoundError {
    pub fn new(details: &str) -> Self {
        NotFoundError {
            details: details.to_string(),
        }
    }
}

impl std::fmt::Display for NotFoundError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Not Found... Error: {}", self.details)
    }
}

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error(transparent)]
    DbErr(#[from] sea_orm::DbErr),

    #[error("Not Found: {0}")]
    NotFound(#[from] NotFoundError),

    #[error(transparent)]
    TaskJoin(#[from] task::JoinError),
}

#[async_trait]
impl AuthnBackend for UserDatabase {
    type User = User;
    type Credentials = Credentials;
    type Error = Error;

    async fn authenticate(
        &self,
        creds: Self::Credentials,
    ) -> Result<Option<Self::User>, Self::Error> {
        let user_found = entity::User::find()
            .filter(entity::user::Column::Username.eq(creds.username))
            .one(&self.connection)
            .await?;

        let user = if let Some(user) = user_found {
            let rear_user: User = user.into();
            Some(rear_user)
        } else {
            None
        };

        // Verifying the password is blocking and potentially slow, so we'll do so via
        // `spawn_blocking`.
        task::spawn_blocking(|| {
            // We're using password-based authentication--this works by comparing our form
            // input with an argon2 password hash.
            Ok(user.filter(|user| verify_password(creds.password, &user.password).is_ok()))
        })
        .await?
    }

    async fn get_user(&self, user_id: &UserId<Self>) -> Result<Option<Self::User>, Self::Error> {
        let user_found = entity::User::find_by_id(*user_id)
            .one(&self.connection)
            .await?;
        if let Some(user) = user_found {
            let rear_user: User = user.into();
            Ok(Some(rear_user))
        } else {
            Ok(None)
        }
    }
}

// We use a type alias for convenience.
//
// Note that we've supplied our concrete backend here.
pub type AuthSession = axum_login::AuthSession<UserDatabase>;
